Exception: Idl::AstNode::InternalError

Inherits:
StandardError
  • Object
show all
Extended by:
T::Sig
Defined in:
lib/idlc/ast.rb

Overview

error that is thrown when the compiler hits an unrecoverable error (that needs fixed!)

Instance Attribute Summary collapse

Instance Method Summary collapse

Constructor Details

#initialize(what) ⇒ InternalError

Returns a new instance of InternalError.



148
149
150
151
152
153
154
155
156
157
# File 'lib/idlc/ast.rb', line 148

def initialize(what)
  super(what)

  @what = what
  @bt = Kernel.caller

  # shift twice to get to the call site of 'internal_error'
  @bt.shift
  @bt.shift
end

Instance Attribute Details

#btObject (readonly)

Returns the value of attribute bt.



145
146
147
# File 'lib/idlc/ast.rb', line 145

def bt
  @bt
end

#whatObject (readonly)

Returns the value of attribute what.



137
138
139
# File 'lib/idlc/ast.rb', line 137

def what
  @what
end