Exception: Idl::AstNode::TypeError
- Inherits:
-
StandardError
- Object
- StandardError
- Idl::AstNode::TypeError
- Extended by:
- T::Sig
- Defined in:
- lib/idlc/ast.rb
Overview
error that is thrown when compilation reveals a type error
Instance Attribute Summary collapse
-
#bt ⇒ Object
readonly
Returns the value of attribute bt.
-
#what ⇒ Object
readonly
Returns the value of attribute what.
Instance Method Summary collapse
-
#initialize(what) ⇒ TypeError
constructor
A new instance of TypeError.
Constructor Details
#initialize(what) ⇒ TypeError
Returns a new instance of TypeError.
119 120 121 122 123 124 125 126 127 128 |
# File 'lib/idlc/ast.rb', line 119 def initialize(what) super(what) @what = what @bt = Kernel.caller # shift twice to get to the call site of 'type_error' @bt.shift @bt.shift end |
Instance Attribute Details
#bt ⇒ Object (readonly)
Returns the value of attribute bt.
115 116 117 |
# File 'lib/idlc/ast.rb', line 115 def bt @bt end |
#what ⇒ Object (readonly)
Returns the value of attribute what.
107 108 109 |
# File 'lib/idlc/ast.rb', line 107 def what @what end |