Class: Idl::NoopAst
- Defined in:
- lib/idlc/ast.rb,
lib/idlc/passes/gen_adoc.rb
Constant Summary
Constants inherited from AstNode
AstNode::Bits1Type, AstNode::Bits32Type, AstNode::Bits64Type, AstNode::BoolType, AstNode::ConstBoolType, AstNode::PossiblyUnknownBits1Type, AstNode::PossiblyUnknownBits32Type, AstNode::PossiblyUnknownBits64Type, AstNode::ReachableFunctionCacheType, AstNode::StringType, AstNode::VoidType
Instance Attribute Summary
Attributes inherited from AstNode
#children, #input, #interval, #parent
Class Method Summary collapse
Instance Method Summary collapse
- #const_eval?(symtab) ⇒ Boolean
-
#execute(symtab) ⇒ void
“execute” the statement by updating the variables in the symbol table.
- #gen_adoc(indent = 0, indent_spaces: 2) ⇒ Object
-
#initialize ⇒ NoopAst
constructor
A new instance of NoopAst.
- #to_h ⇒ Object
- #to_idl ⇒ Object
-
#type_check(symtab, strict:) ⇒ void
type check this node and all children.
Methods inherited from AstNode
#always_terminates?, #declaration?, #executable?, extract_base_var_name, #find_ancestor, #find_dst_registers, #find_referenced_csrs, #find_src_registers, #freeze_tree, #gen_option_adoc, #input_file, input_from_source_yaml, #inspect, #internal_error, interval_from_source_yaml, #lineno, #lines_around, #nullify_assignments, #pass_find_return_values, #path, #print_ast, #prune, #reachable_exceptions, #reachable_functions, #set_input_file, #set_input_file_unless_already_set, #source_line_file_offsets, #source_starting_offset, #source_yaml, #starting_line, #text_value, #truncation_warn, #type_error, #unindent, value_else, #value_else, value_error, #value_error, value_try, #value_try, write_back_nested
Constructor Details
#initialize ⇒ NoopAst
Returns a new instance of NoopAst.
6547 6548 6549 |
# File 'lib/idlc/ast.rb', line 6547 def initialize super("", 0...0, EMPTY_ARRAY) end |
Class Method Details
Instance Method Details
#const_eval?(symtab) ⇒ Boolean
6545 |
# File 'lib/idlc/ast.rb', line 6545 def const_eval?(symtab) = true |
#execute(symtab) ⇒ void
This method returns an undefined value.
“execute” the statement by updating the variables in the symbol table
6555 |
# File 'lib/idlc/ast.rb', line 6555 def execute(symtab); end |
#gen_adoc(indent = 0, indent_spaces: 2) ⇒ Object
17 |
# File 'lib/idlc/passes/gen_adoc.rb', line 17 def gen_adoc(indent = 0, indent_spaces: 2) = "" |
#to_h ⇒ Object
6562 6563 6564 6565 |
# File 'lib/idlc/ast.rb', line 6562 def to_h = { "kind" => "noop_expr", "source" => source_yaml } |
#to_idl ⇒ Object
6559 |
# File 'lib/idlc/ast.rb', line 6559 def to_idl = "" |
#type_check(symtab, strict:) ⇒ void
This method returns an undefined value.
type check this node and all children
Calls to #type and/or #value may depend on type_check being called first with the same symtab. If not, those functions may raise an AstNode::InternalError
6552 |
# File 'lib/idlc/ast.rb', line 6552 def type_check(symtab, strict:); end |